/* alamo.c by K. Spoon <kspoon@rackspace.com>
 * Copyright Rackspace (C) 2001
 * 
 * Derived from knark.c... see comments below
 *
 * This is a kernel module (should work for 2.2.X kernels) and is based on the
 * knark.c kernel root kit by Creed.
 *
 * Essentially, this module uses knark's trick of trojaning kernel syscall's
 * except this time we're putting in the real deal Holyfield syscalls.  The
 * syscalls() you'll find below were ripped out of the 2.2.14 kernel.
 *
 * What this does is expose everything that knark tried to hide, which should
 * give harried sysadmins a fighting chance to find the rootkit and other 
 * unfriendly stuff on their system.
 *
 * See http://www.rackspace.com/alamo/ for more details
 */


#define __KERNEL_SYSCALLS__
#include <linux/modversions.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/sched.h>
#include <linux/socket.h>
#include <linux/smp_lock.h>
#include <linux/stat.h>
#include <linux/dirent.h>
#include <linux/fs.h>
#include <linux/if.h>
#include <linux/modversions.h>
#include <linux/malloc.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <sys/syscall.h>
#include <net/protocol.h>
#include <net/udp.h>
#include <net/icmp.h>
#include <sys/syscall.h>

#include <linux/dirent.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/errno.h>

#define ALAMO_VERSION "v0.1"
#define MODULE_NAME "alamo"

#define NAME_OFFSET(de) ((int) ((de)->d_name - (char *) (de)))
#define ROUND_UP(x) (((x)+sizeof(long)-1) & ~(sizeof(long)-1))

struct getdents_callback {
//	struct linux_dirent * current_dir;
//	struct linux_dirent * previous;
	struct dirent * previous;
	struct dirent * current_dir;
	int count;
	int error;
};

extern void *sys_call_table[];


static inline _syscall3(int, getdents, uint, fd, void *, dirp, uint, count)
static inline _syscall2(int, kill, int, pid, int, sig)
static inline _syscall3(int, ioctl, int, fd, int, cmd, long, arg)
// Yanking fork, since that looks like it's going to be naaaaasty 
//static inline _syscall1(int, fork, int, regs)
//static inline _syscall1(int, clone, int, regs)
//static inline _syscall2(int, settimeofday, struct timeval *, tv, struct timezone *, tz) 

int (*original_getdents)(unsigned int, void *, unsigned int);
int (*original_kill)(int, int);
int (*original_read)(unsigned int, char *, size_t);
int (*original_ioctl)(int, int, long);
// Yanking fork, since that looks like it's going to be naaaaasty 
//int (*original_fork)(struct pt_regs);
//int (*original_clone)(struct pt_regs);
int (*original_execve)(struct pt_regs);
//int (*original_settimeofday)(struct timeval *, struct timezone *);

/* Functions that the syscall()s rely on */
// alamo_do_fork
// alamo_kill_something_info
// alamo_locks_mandatory_area
// int alamo_do_sys_settimeofday( struct timeval *tv, struct timezone *tz );

extern int file_ioctl( struct file *filp, unsigned int cmd, unsigned long arg);
extern int do_sys_settimeofday( struct timeval *tv, struct timezone *tz );
extern int filldir( void *, const char *, int, off_t, ino_t);
extern void fput(struct file *file);

extern inline struct file * fcheck(unsigned int fd)
{
  struct file * file = NULL;

  if (fd < current->files->max_fds)
    file = current->files->fd[fd];
  return file;
}

extern inline struct file * fget(unsigned int fd)
{
  struct file * file = fcheck(fd);

  if (file)
    file->f_count++;
  return file;
}

static int alamo_filldir(void * __buf, const char * name, int namlen, off_t offset, ino_t ino)
{
//  struct linux_dirent * dirent;
  struct dirent * dirent;
  struct getdents_callback * buf = (struct getdents_callback *) __buf;
  int reclen = ROUND_UP(NAME_OFFSET(dirent) + namlen + 1);

  buf->error = -EINVAL; /* only used if we fail.. */
  if (reclen > buf->count)
    return -EINVAL;
  dirent = buf->previous;
  if (dirent)
    put_user(offset, &dirent->d_off);
  dirent = buf->current_dir;
  buf->previous = dirent;
  put_user(ino, &dirent->d_ino);
  put_user(reclen, &dirent->d_reclen);
  copy_to_user(dirent->d_name, name, namlen);
  put_user(0, dirent->d_name + namlen);
  ((char *) dirent) += reclen;
  buf->current_dir = dirent;
  buf->count -= reclen;
  return 0;
}

static int alamo_file_ioctl(struct file *filp,unsigned int cmd,unsigned long arg)
{
  int error;
  int block;
  struct inode * inode = filp->f_dentry->d_inode;

  switch (cmd) {
    case FIBMAP:
      if (!capable(CAP_SYS_RAWIO))
        return -EPERM;
      if (inode->i_op == NULL)
        return -EBADF;
          if (inode->i_op->bmap == NULL)
        return -EINVAL;
      if ((error = get_user(block, (int *) arg)) != 0)
        return error;
      block = inode->i_op->bmap(inode,block);
      return put_user(block, (int *) arg);
    case FIGETBSZ:
      if (inode->i_sb == NULL)
        return -EBADF;
      return put_user(inode->i_sb->s_blocksize, (int *) arg);
    case FIONREAD:
      return put_user(inode->i_size - filp->f_pos, (int *) arg);
  }
  if (filp->f_op && filp->f_op->ioctl)
    return filp->f_op->ioctl(inode, filp, cmd, arg);
  return -ENOTTY;
}

int alamo_getdents(unsigned int fd, void* dirp, unsigned int count)
{
	struct file * file;
	struct dentry * dentry;
	struct inode * inode;
//	struct linux_dirent * lastdirent;
	struct dirent * lastdirent;
	struct getdents_callback buf;
	int error;

	lock_kernel();
	error = -EBADF;
	file = fget(fd);
	if (!file)
		goto out;

	dentry = file->f_dentry;
	if (!dentry)
		goto out_putf;
	inode = dentry->d_inode;
	if (!inode)
		goto out_putf;
//	buf.current_dir = (struct linux_dirent *) dirp;
	buf.current_dir = (struct dirent *) dirp;
	buf.previous = NULL;
	buf.count = count;
	buf.error = 0;
	error = -ENOTDIR;

	if (!file->f_op || !file->f_op->readdir)
		goto out_putf;
										        /*
	 * Get the inode's semaphore to prevent changes
	 * to the directory while we read it.
	 */
	down(&inode->i_sem);
	error = file->f_op->readdir(file, &buf, alamo_filldir);

	up(&inode->i_sem);
	if (error < 0)
		goto out_putf;
	error = buf.error;
	lastdirent =  buf.previous;
	if (lastdirent) {
		/* put_user((int) file->f_pos, (void*) &lastdirent->d_off); */
		error = count - buf.count;
	}

out_putf:
	fput(file);
out:
	unlock_kernel();
	return error;
}

int alamo_kill_something_info(int sig, struct siginfo *info, int pid)
{
  if (!pid) {
    return kill_pg_info(sig, info, current->pgrp);
  } else if (pid == -1) {
    int retval = 0, count = 0;
    struct task_struct * p;

    read_lock(&tasklist_lock);
    for_each_task(p) {
      if (p->pid > 1 && p != current) {
        int err = send_sig_info(sig, info, p);
        ++count;
        if (err != -EPERM)
          retval = err;
      }
    }
    read_unlock(&tasklist_lock);
    return count ? retval : -ESRCH;
  } else if (pid < 0) {
    return kill_pg_info(sig, info, -pid);
  } else {
    return kill_proc_info(sig, info, pid);
  }
}

int alamo_kill(pid_t pid, int sig)
{
	struct siginfo info;
	memset(&info, 0, sizeof(info));
	info.si_signo = sig;
	info.si_errno = 0;
	info.si_code = SI_USER;
	info.si_pid = current->pid;
	info.si_uid = current->uid;

	return alamo_kill_something_info(sig, &info, pid);
}

// Yanking read, because it looks like it's bringing in a symbol with a huge
// func behind it.
//int alamo_read(int fd, char *buf, size_t count)
//{
//	ssize_t ret;
//	struct file * file;
//	ssize_t (*read)(struct file *, char *, size_t, loff_t *);
//
//	lock_kernel();
//	ret = -EBADF;
//	file = fget(fd);
//	if (!file)
//		goto bad_file;
//	if (!(file->f_mode & FMODE_READ))
//		goto out;
//	ret = locks_verify_area(FLOCK_VERIFY_READ, file->f_dentry->d_inode,
//	                        file, file->f_pos, count);
//
//	if (ret)
//		goto out;
//
//	ret = -EINVAL;
//
//	if (!file->f_op || !(read = file->f_op->read))
//		goto out;
//	ret = read(file, buf, count, &file->f_pos);
//out:
//	fput(file);
//bad_file:
//	unlock_kernel();
//	return ret;
//}

int alamo_ioctl(int fd, int cmd, long arg)
{
	struct file * filp;
	unsigned int flag;
	int on, error = -EBADF;

	lock_kernel();
	filp = fget(fd);
	if (!filp)
		goto out;
	error = 0;
	switch (cmd) {
		case FIOCLEX:
			FD_SET(fd, current->files->close_on_exec);
			break;
		case FIONCLEX:
			FD_CLR(fd, current->files->close_on_exec);
			break;
		case FIONBIO:
			if ((error = get_user(on, (int *)arg)) != 0)
				break;
			flag = O_NONBLOCK;
#ifdef __sparc__
			/* SunOS compatibility item. */
			if(O_NONBLOCK != O_NDELAY)
				flag |= O_NDELAY;
#endif
			if (on)
				filp->f_flags |= flag;
			else
				filp->f_flags &= ~flag;
			break;
		case FIOASYNC:
			if ((error = get_user(on, (int *)arg)) != 0)
				break;
			flag = on ? FASYNC : 0;

			/* Did FASYNC state change ? */
			if ((flag ^ filp->f_flags) & FASYNC) {
				if (filp->f_op && filp->f_op->fasync)
					filp->f_op->fasync(fd, filp, on);
			}

			if (on)
				filp->f_flags |= FASYNC;
			else
				filp->f_flags &= ~FASYNC;
			break;

		default:
			error = -ENOTTY;
			if (!filp->f_dentry || !filp->f_dentry->d_inode)
				error = -ENOENT;
			else if (S_ISREG(filp->f_dentry->d_inode->i_mode))
				error = alamo_file_ioctl(filp, cmd, arg);
			else if (filp->f_op && filp->f_op->ioctl)
				error = filp->f_op->ioctl(filp->f_dentry->d_inode, filp, cmd, arg);
		}
	        fput(filp);
out:
		unlock_kernel();
		return error;
}

// Yanking fork, since that looks like it's going to be naaaaasty 
//int alamo_fork(struct pt_regs regs)
//{
//	return do_fork( SIGCHLD, regs.esp, &regs );
//}

// Yanking fork, since that looks like it's going to be naaaaasty 
//int alamo_clone(struct pt_regs regs)
//{
//	        unsigned long clone_flags;
//		unsigned long newsp;
//
//	        clone_flags = regs.ebx;
//	        newsp = regs.ecx;
//	        if (!newsp)
//			newsp = regs.esp;
//		return do_fork(clone_flags, newsp, &regs);
//}

/* yanking due to the fact that I'm going to have to import big functions
int alamo_settimeofday(struct timeval *tv, struct timezone *tz)
{
	struct timeval  new_tv;
	struct timezone new_tz;
	int retval;

	if (tv) {
		if (copy_from_user(&new_tv, tv, sizeof(*tv)))
			return -EFAULT;
	}

	if (tz) {
		if (copy_from_user(&new_tz, tz, sizeof(*tz)))
			return -EFAULT;
	}

	retval =  do_sys_settimeofday(tv ? &new_tv : NULL, tz ? &new_tz : NULL);
	return( retval );
}
*/
    
int alamo_execve(struct pt_regs regs)
{
	int error;
	char * filename;

	lock_kernel();
	filename = getname((char *) regs.ebx);
	error = PTR_ERR(filename);
	if (IS_ERR(filename))
		goto out;
	error=do_execve(filename, (char **) regs.ecx,(char **) regs.edx, &regs);

	if (error == 0)
		current->flags &= ~PF_DTRACE;
	putname(filename);

out:
	unlock_kernel();
	return error;
}

int init_module(void)
{
     original_getdents = sys_call_table[SYS_getdents];
     sys_call_table[SYS_getdents] =  alamo_getdents;
    
     original_kill =  sys_call_table[SYS_kill];
     sys_call_table[SYS_kill] =  alamo_kill;
    
// Yanking read, because it looks like it's bringing in a symbol with a huge
// func behind it.
//     original_read =  sys_call_table[SYS_read];
//     sys_call_table[SYS_read] =  alamo_read;
    
     original_ioctl =  sys_call_table[SYS_ioctl];
     sys_call_table[SYS_ioctl] =  alamo_ioctl;
    
// Yanking fork, since that looks like it's going to be naaaaasty 
//     original_fork =  sys_call_table[SYS_fork];
//     sys_call_table[SYS_fork] =  alamo_fork;
    
//     original_clone =  sys_call_table[SYS_clone];
//     sys_call_table[SYS_clone] =  alamo_clone;
    
//  yanking... brings in too much crap
//     original_settimeofday =  sys_call_table[SYS_settimeofday];
//     sys_call_table[SYS_settimeofday] =  alamo_settimeofday;

     original_execve =  sys_call_table[SYS_execve];
     sys_call_table[SYS_execve] =  alamo_execve;

    return 0;
}


void cleanup_module(void)
{
     sys_call_table[SYS_getdents] = original_getdents;
     sys_call_table[SYS_kill] = original_kill;
// Yanking because needs a symbol with a big func behind it
//     sys_call_table[SYS_read] = original_read;
     sys_call_table[SYS_ioctl] = original_ioctl;
// Yanking fork, since that looks like it's going to be naaaaasty 
//     sys_call_table[SYS_fork] = original_fork;
//     sys_call_table[SYS_clone] = original_clone;
//     sys_call_table[SYS_settimeofday] = original_settimeofday;
     sys_call_table[SYS_execve] = original_execve;
}
